\ FILE:         sievmain.frt
\ LANGUAGE : ANSI Forth
\ COPYRIGHT :  Albert van der Horst FIG Chapter Holland
\ This program may be copied and distributed freely as is.
\ This program may only be modified
\ 1. with purpose :
\     configuration  or
\     to comment out parts of the program where ANSI words
\      are not available or non-ANSI words are already available  or
\     to improve the algorithm  or
\     to improve the documentation  or
\     to put the code in blocks  or
\     to port it to older (non ANS) systems
\ and
\ 2. Clear indication of changes and the person/company responsible for it.
\ Expressly forbidden is
\ 1. gratitious de-ansification, like use of non-standard comment symbols
\ 2. lower casing
\ 3. comment stripping
\ 4. addition of system-specific words like source control tools such as
\     MARKER look-alikes
\
\ PROJECT : TWIN PRIMES
\ CATEGORY : EXAMPLE
\ AUTHOR :  Albert van der Horst albert@spenarnc.xs4all.nl avdhorst@doge.nl
\ DESCRIPTION: Generates tables of prime numbers
\ VERSION : 1.0 bases on ssieve 3.3 A
\ LAST CHANGE : 1998 jan 12 AH Reading from disk if possible
\ CREATED : 1996 oct 13

\ Super sieve, up to at least 10^9 by Albert van der HORST, HCC Forth gg

\ The well known sieve of Eratosthenes implemented in FORTH.
\ This implementation will spit out all primes up till at least the
\ precision of the square of MAXINT, so ca. 10^18 on a 32-bit machine. It
\ is done in batches, i.e. it is possible to calculate any range. Also it
\ has some beautiful printer arrangements to print your own books with
\ prime tables.

\ Some theoretisation.
\ This problem abounds with structure clashes. The number of primes that
\ fit on a page or line, the number {THOUSAND} of primes that in the user view
\ belong together, and the SIEVE-SIZE of a batch, and the number of primes that
\ fit in a byte, all clash.

    NEEDS -crc
\ Comment out if not needed/present
    MARKER ssieve

\ PART 3: CALCULATION CONFIGURATION

\ WARNING: SIEVE-SIZE must not be negative (unsigned and > MAX-N)
\           Disasters will happen in ELIMINATE/30
\ This will hold 3.3*10^7 primes sufficient for the world
\ twin prime record of 10^15
      1100000 CONSTANT SIEVE-SIZE
\       110000 CONSTANT SIEVE-SIZE     \ testing: will do until 10^13
    3 CONSTANT N-DIGITS \ Number of digits of mantissa
 1000 CONSTANT THOUSAND \ Must agree with N-DIGITS


\ PART 4: DATA STRUCTURES AND ELEMENTARY MANIPULATIONS

\  Both of the arrays contain bit tables, where each byte
\  represents a range i..i+29 where i is a multiple of 30.
\  In each 30 consequitive numbers precisely 8 have no divider
\  in common with 30.
\  Only these are the numbers that could be prime;
\  obviously numbers like 30*i+5 and 30*i+21 are no candidates.
\  SIEVE is reused, when more than "SIEVE-SIZE"*30 numbers are to be
\  investigated.
\  In that case we still need the single precision primes
\  to elimate numbers. Therefore the table PRIMES is always kept.
\  SIEVE is always used as a whole. Not all bits may be inspected

CREATE SIEVE  SIEVE-SIZE ALLOT        \ Batch of candidates primes
0 VARIABLE SIEVE-CRC    \ Crc check over the SIEVE-SIZE bytes of the SIEVE
0 VALUE END-SIEVE    \ Where to stop sieving
VARIABLE SIEVE-FILLED    \ "The table prime is already filled"
FALSE SIEVE-FILLED !     \ Not yet.

\ Give the meaning of bits in a sieve table e.g. bit 7 of a byte
\ represents a 30-fold +29
CREATE BIT-VALUE
\   0     1     2     3     4     5     6     7
    1 C,  7 C, 11 C, 13 C, 17 C, 19 C, 23 C, 29 C,

( ************************)  HEX ( ************************)

\  The array S-MASK is such that for the number 30*i+j the j-th mask
\  gives the bit {in byte i} of that number, but 0 for non-candidates.
\  Used for Clearing bits.
CREATE S-MASK
\   0     1     2     3     4     5     6     7     8     9
   00 C, 01 C, 00 C, 00 C, 00 C, 00 C, 00 C, 02 C, 00 C, 00 C,
\  10    11    12    13    14    15    16    17    18    19
   00 C, 04 C, 00 C, 08 C, 00 C, 00 C, 00 C, 10 C, 00 C, 20 C,
\  20    21    22    23    24    25    26    27    28    29
   00 C, 00 C, 00 C, 40 C, 00 C, 00 C, 00 C, 00 C, 00 C, 80 C,

\ Compile the inverted mask of U.
: I,    ( u U -- )
    INVERT C,
;

\ Like C-MASK but inverted. Used for Setting bits.
CREATE C-MASK
\   0     1     2     3     4     5     6     7     8     9
   00 I, 01 I, 00 I, 00 I, 00 I, 00 I, 00 I, 02 I, 00 I, 00 I,
\  10    11    12    13    14    15    16    17    18    19
   00 I, 04 I, 00 I, 08 I, 00 I, 00 I, 00 I, 10 I, 00 I, 20 I,
\  20    21    22    23    24    25    26    27    28    29
   00 I, 00 I, 00 I, 40 I, 00 I, 00 I, 00 I, 00 I, 00 I, 80 I,

\ Preset SIEVE to all primes.
: INIT-T   SIEVE END-SIEVE OVER - 0FF FILL ;

( ************************)  DECIMAL ( ************************)


\  In the following if the double number happens to be a
\  2,3 or 5 fold, one of the dummy masks is selected and
\  nothing changes; for TEST-B a FALSE is returned

\ Clears the bit of the prime candidate with (number) offset D in SIEVE
: CLEAR-B ( d D --  )
     30 UM/MOD                   \ Leave mask number and index
     SIEVE +                    \ Address in table
     SWAP C-MASK + C@           \ Get mask
     OVER C@ AND SWAP C!        \ Now clear that bit
;

\ Sets the bit of the prime candidate with (number) offset D in SIEVE
: SET-B ( d --  sets the bit of the prime candidate )
     30 UM/MOD                   \ Leave mask number and index
     SIEVE +                    \ Address in table
     SWAP S-MASK + C@           \ Get mask
     OVER C@ OR SWAP C!         \ Now clear that bit
;

\ Returns the bit of the prime candidate with (number) offset D in SIEVE
\ as FLAG.
\ WARNING: the result can be used by IF , but has not all bits set if true
: TEST-B ( d D -- n FLAG )
     30 UM/MOD                  \ Leave mask number and index
     SIEVE + C@                 \ Byte from table
     SWAP S-MASK + C@           \ Get mask
     AND                        \ Result: =0 or #0
;

\ Returns whether U is prime according to our small table SIEVE as FLAG.
\ It errs on the safe side, by saying yes if not present in table.
\ WARNING: the result can be used by IF , but has not all bits set if true
: PRIME? ( u U --  n FLAG )
     U>D 30 UM/MOD               \ Leave mask number and index
     DUP SIEVE-SIZE < 0= IF 2DROP TRUE EXIT THEN
     SIEVE + C@                \ Byte from table
     SWAP S-MASK + C@           \ Get mask
     AND                        \ Result: =0 or #0
;

\ Clears the bits that are zero in MASK at address A
: CLEAR-BIT ( u MASK,addr A --  )
     SWAP OVER C@               \ Old value
     AND SWAP C!                \ Now clear that bit
;

VARIABLE PRIME
\ Eliminate 1/30 of the multiples of prime in the SIEVE
\ i.e. all the multiples at BYTE-INDEX +x*PRIME
\ If the PRIME is large a wrap over all memory again into the table
\ might occurs.
\ If both the PRIME and the length of the table SIEVE
\ are smaller than MAX-U, this cannot occur.
\ WARNING: Uses the global value END-SIEVE that is sometimes manipulated
: ELIMINATE/30       ( u MASK, n BYTE-INDEX -- )
   \ Skip if divisible by 2,3,5
   OVER [ HEX ] 0FF [ DECIMAL ] <> IF
       SIEVE +              \ Change to address in table SIEVE
       BEGIN
          DUP SIEVE END-SIEVE WITHIN
       WHILE
          2DUP CLEAR-BIT
          PRIME @ +          \  next addres to be cleared with the same MASK
       REPEAT
   THEN
   2DROP ( address & mask )
;

\ Calculate from the number index N-INDEX in table SIEVE two offsets: the
\ byte in the table where that number is represented, and the bit offset
\ within the table of masks
\ It is to be seen whether this is a good factorization
: SPLIT-INDEX   ( d N-INDEX -- n BYTE-INDEX, u BIT-INDEX )
      30 UM/MOD
;

\ The double at the low end of the SIEVE plus START is divisible by PRIME.
\ This is sufficient knowledge to eliminate all PRIME-fold's
\ from the table. For each possible mask there is a scan
\ through the table, for START, START+PRIME, START+2*PRIME etc.
\ NOTE : It is tricky, but PRIME and START count numbers in the table,
\ not bytes, not bits. That's why START is changed into a DOUBLE.
\ This once was a pretty bug. { Thanks Lennart! }
\ For the limitation of ELIMINATE/30 PRIME must be < MAX-N
: ELIMINATE ( n PRIME, u START -- )
    SWAP PRIME !
\     PRIME ?
    U>D
    30 0 DO
      2DUP
      SPLIT-INDEX               \ Get index in SIEVE for start, and bit index
      SWAP C-MASK + C@  SWAP    \ Turn bit index into mask
      ELIMINATE/30              \ Eliminate all with the same mask
      PRIME @ U>D D+            \ The next starting point
    LOOP
    2DROP ( START)
;


\ PART 5: INPUT OUTPUT

2VARIABLE COUNTER

\ COUNTER can be checked against the following table
\    X              PI(X) : Number of primes <= X :
\         1,000                168      Confirmed from litterature
\        10,000              1,229        "
\       100,000              9,592        "
\       900,000             71,274        "
\     1,000,000             78,498        "
\    10,000,000            664,579        "
\   100,000,000          5,761,455        "
\ 1,000,000,000         50,847,534        "
\ 2G145 .. 2G146            46,754      By this program
\ 2G .. 2G1 .            4,864,551        "
\ 4,000,000,000        189,961,812        "



\ PART 6: SIEVING, at last

\ Compares the crc of the sieve with the crc stored in SIEVE-CRC.
\ Returns "crc is okay". This fails e.g. in case the SIEVE has been
\ corrupted during the running of the program.
: CHECK-SIEVE-CRC  ( -- int OKAY)
    SIEVE SIEVE-SIZE OVER + CrcOfArea    SIEVE-CRC @   =
;


\ Try to read the table from the file with the name PRIMES.TAB
\ If this table doesnot contain correct data, bad luck. after this piece of
\ code has been tested, we are no longer interested why this fails, so
\ replace with straight THROWS
: TRY-READ              ( -- )
    S" PRIMES.TAB" R/O BIN OPEN-FILE THROW
    LOCAL FILE
    SIEVE SIEVE-SIZE FILE READ-FILE THROW
    SIEVE-SIZE <> IF FILE CLOSE-FILE DROP TRUE ABORT" Too small" THEN
    SIEVE-CRC 1 CELLS FILE READ-FILE THROW
    1 CELLS <> IF FILE CLOSE-FILE DROP TRUE ABORT" No CRC" THEN
    ." Checking CRC"
    CHECK-SIEVE-CRC  0= IF
        FILE CLOSE-FILE DROP TRUE ABORT" CRC wrong"
    THEN
    FILE CLOSE-FILE THROW
    TRUE SIEVE-FILLED !
    ." Primes read succesfully"
;

\ Try to write the table to a file with the name PRIMES.TAB
\ We are absolutely not interested why this fails, so straight THROWS
: TRY-WRITE              ( -- )
    S" PRIMES.TAB" R/W BIN CREATE-FILE IF
        ( fid) DROP S" PRIMES.TAB" R/W BIN OPEN-FILE THROW
    THEN
    ( fid) LOCAL FILE
    SIEVE SIEVE-SIZE FILE WRITE-FILE THROW
    SIEVE SIEVE-SIZE OVER + CrcOfArea    SIEVE-CRC !
    SIEVE-CRC 1 CELLS FILE WRITE-FILE THROW
    FILE CLOSE-FILE THROW
    ." Primes written succesfully"
;

\ Sieve the first batch of SIEVE. Remember up to 30*"SIEVE-SIZE" primes in the
\ compact prime-table PRIMES.
: FILL-PRIMES
     SIEVE-FILLED @ 0=
     ['] TRY-READ CATCH 0<>
     AND IF              \ We need to do this just once
         SIEVE SIEVE-SIZE + TO END-SIEVE
         ." Sieving"
         INIT-T
         1. CLEAR-B                     \ 1 is not prime.
         7
         BEGIN
              DUP S>D TEST-B IF
                DUP DUP DUP UM* D>S ELIMINATE        \ Pass prime, 2*prime as parameters
              THEN
              2 +
              DUP DUP UM*   SIEVE-SIZE 30 UM*   D< WHILE
         REPEAT DROP CR
         TRUE SIEVE-FILLED !
         ['] TRY-WRITE CATCH 0<> IF CR ." Writing prime table failed" THEN
     THEN
;

\ SIEVE has now been filled up till SIEVE-SIZE
